/* glom.c: builds an argument list out of words, variables, etc. */

#include "rc.h"

#include <errno.h>
#include <termios.h>

static List *backq(Node *, Node *);
static List *bqinput(List *, int, bool);
static List *count(List *);
static List *mkcmdarg(Node *);

static bool allow_nil_first_var;
Rq *redirq = NULL;

List *word(char *w, char *m) {
	List *s = NULL;
	if (w != NULL) {
		s = nnew(List);
		s->w = w;
		s->m = m;
		s->n = NULL;
	}
	return s;
}

/*
   Append list s2 to list s1 by copying s1 and making the new copy
   point at s2.
*/

List *append(List *s1, List *s2) {
	List *r, *top;
	if (s1 == NULL)
		return s2;
	if (s2 == NULL)
		return s1;
	for (r = top = nnew(List); 1; r = r->n = nnew(List)) {
		r->w = s1->w;
		r->m = s1->m;
		if ((s1 = s1->n) == NULL)
			break;
	}
	r->n = s2;
	return top;
}

/* Append list s2 to s1 without any copying or allocation. */
List *append_noalloc(List *s1, List *s2) {
	if (s1 == NULL)
		return s2;
	if (s2 == NULL)
		return s1;
	List *r = s1;
	while (r->n != NULL)
		r = r->n;
	r->n = s2;
	return s1;
}

List *concat(List *s1, List *s2) {
	int n1, n2;
	List *r, *top;
	if (s1 == NULL)
		return s2;
	if (s2 == NULL)
		return s1;
	if ((n1 = listnel(s1)) != (n2 = listnel(s2)) && n1 != 1 && n2 != 1)
		rc_error("bad concatenation");
	for (r = top = nnew(List); 1; r = r->n = nnew(List)) {
		size_t x = strlen(s1->w);
		size_t y = strlen(s2->w);
		size_t z = x + y + 1;
		r->w = nalloc(z);
		memcpy(r->w, s1->w, x);
		memcpy(r->w+x, s2->w, y+1);
		if (s1->m == NULL && s2->m == NULL) {
			r->m = NULL;
		} else {
			r->m = nalloc(z);
			if (s1->m == NULL)
				memzero(r->m, x);
			else
				memcpy(r->m, s1->m, x);
			if (s2->m == NULL)
				memzero(&r->m[x], y);
			else
				memcpy(&r->m[x], s2->m, y);
			r->m[z] = 0;
		}
		if (n1 > 1)
			s1 = s1->n;
		if (n2 > 1)
			s2 = s2->n;
		if (s1 == NULL || s2 == NULL || (n1 == 1 && n2 == 1))
			break;
	}
	r->n = NULL;
	return top;
}

void bad_subscript(void) {
	rc_error("bad subscript");
}

int parse_index(char *word) {
	int i;
	i = a2u(word);
	if (i < 1)
		bad_subscript();
	return i;
}

/* Parse a variable subscript, whether is a range ($from-$to) or an index. */
void parse_subscript(char *word, int n, int *from, int *to) {
	char *dash = strchr(word, '-');

	/* No '-' found, this is just an index number. */
	if (dash == NULL) {
		*from = *to = parse_index(word);
		return;
	}
	/* '-' found at the start. We do not accept negative numbers. */
	if (dash == word)
		bad_subscript();
	/* Make sure there is only 1 '-'. */
	if (strchr(dash+1, '-') != NULL)
		bad_subscript();
	
	*dash = '\0';
	*from = parse_index(word);
	char *t = dash + 1;
	if (*t == '\0') {
		*to = MAX(n, *from);
	} else {
		*to = parse_index(t);
		if (*to > n) *to = n;
		if (*to < *from) bad_subscript();
	}
	*dash = '-';
}

List *varsub(List *var, List *subs) {
	List *r, *top;
	int from, to;

	int n = listnel(var);

	for (top = r = NULL; subs != NULL; subs = subs->n) {
		parse_subscript(subs->w, n, &from, &to);
		if (from > n)
			continue;

		List *sub = var;
		int range = to - from + 1;
		while (--from)
			sub = sub->n; /* loop until sub == var(from) */

		while (range--) {
			if (top == NULL)
				top = r = nnew(List);
			else
				r = r->n = nnew(List);
			r->w = sub->w;
			r->m = sub->m;
			sub = sub->n;
		}
	}
	if (top != NULL)
		r->n = NULL;
	return top;
}

List *flatten(List *s) {
	List *r, *p, *ofs;
	size_t size, step, sl;
	char *sep, *f;
	if (s == NULL || s->n == NULL)
		return s;
	ofs = varlookup_safe("ofs");
	if (ofs != NULL && ofs->w != NULL)
		sep = ofs->w;
	else
		sep = " ";
	sl = strlen(sep);
	for (size = 0, p = s; p != NULL; p = p->n)
		size += strlen(p->w) + (p->n == NULL ? 0 : sl);
	r = nnew(List);
	f = r->w = nalloc(size + 1);
	r->m = NULL; /* flattened lists come from variables, so no meta */
	r->n = NULL;
	step = strlen(s->w);
	memcpy(f, s->w, step);
	f += step;
	do {
		memcpy(f, sep, sl);
		f += sl;
		s = s->n;
		step = strlen(s->w);
		memcpy(f, s->w, step);
		f += step;
	} while (s->n != NULL);
	*f = '\0';
	return r;
}

static List *count(List *l) {
	List *s = nnew(List);
	s->w = nprint("%d", listnel(l));
	s->n = NULL;
	s->m = NULL;
	return s;
}

void assign(List *s1, List *s2, bool stack) {
	List *val = s2;
	if (s1 == NULL)
		rc_error("null variable name");
	if (s1->n != NULL)
		rc_error("multi-word variable name");
	if (*s1->w == '\0')
		rc_error("zero-length variable name");
	if (a2u(s1->w) != -1)
		rc_error("numeric variable name");
	if (strchr(s1->w, '=') != NULL)
		rc_error("'=' in variable name");
	if (*s1->w == '*' && s1->w[1] == '\0')
		val = append(varlookup("0"), s2); /* preserve $0 when * is assigned explicitly */
	if (s2 != NULL || stack) {
		if (dashex)
			prettyprint_var(2, s1->w, val);
		varassign(s1->w, val, stack);
		alias(s1->w, varlookup(s1->w), stack);
	} else {
		if (dashex)
			prettyprint_var(2, s1->w, NULL);
		varrm(s1->w, stack);
	}
}

/* Try to optimize cases like: x=($x ...) by using varappend()
   instead of varassign(), which avoids duplicating $x.
   Return true on success. Else, you probably need to use assign(). */
bool tryappend(List *name, Node *n) {

	/* We do not need to validate "name" like in assign(),
	   since varappend() will return false if the variable
	   has not already been created with assign(). */

	if (n == NULL || n->type != nLappend)
		return false;
	
	/* Find the deepest nLappend node. */
	Node *an = n;
	while (an->u[0].p->type == nLappend)
		an = an->u[0].p;

	/* Check if the first variable has the same name. */
	Node *var = an->u[0].p;
	if (var->type != nVar)
		return false;
	char *vname = var->u[0].p->u[0].s;
	if (vname == NULL || !streq(vname, name->w))
		return false;
	
	/* Append operation detected! */
	/* Do another traversal to force left-to-right evaluation. */
	List *head, *tail = NULL;
	while (1) {
		head = glom(n->u[1].p);
		if (head != NULL)
			tail = append(head, tail);
		if (n == an)
			break;
		n = n->u[0].p;
	}

	/* Do the glob on the whole list to get sorted results. */
	return varappend(name->w, glob(tail));
}

/*
   The following two functions are by the courtesy of Paul Haahr,
   who could not stand the incompetence of my own backquote implementation.
*/

#define BUFSIZE	((size_t) 1000)

static List *bqinput(List *ifs, int fd, bool trim) {
	char *end, *bufend, *s;
	List *r, *top, *prev;
	size_t remain, bufsize;
	char isifs[256];
	ssize_t n;
	int state; /* a simple FSA is used to read in data */

	memzero(isifs, sizeof isifs);
	for (isifs['\0'] = true; ifs != NULL; ifs = ifs->n)
		for (s = ifs->w; *s != '\0'; s++)
			isifs[*(unsigned char *)s] = true;
	remain = bufsize = BUFSIZE;
	top = r = nnew(List);
	r->w = end = nalloc(bufsize + 1);
	r->m = NULL;
	state = 0;
	prev = NULL;

	while (1) {
		if (remain == 0) { /* is the string bigger than the buffer? */
			size_t m = end - r->w;
			char *buf;
			while (bufsize < m + BUFSIZE)
				bufsize *= 2;
			buf = nalloc(bufsize + 1);
			memcpy(buf, r->w, m);
			r->w = buf;
			end = &buf[m];
			remain = bufsize - m;
		}
		if ((n = rc_read(fd, end, remain)) <= 0) {
			if (n == 0)
	/* break */		break;
			else if (errno == EINTR)
				return NULL; /* interrupted, wait for subproc */
			else {
				uerror("backquote read");
				rc_error(NULL);
			}
		}
		remain -= n;
		for (bufend = &end[n]; end < bufend; end++)
			if (state == 0) {
				if (!isifs[*(unsigned char *)end]) {
					state = 1;
					r->w = end;
					r->m = NULL;
				}
			} else {
				if (isifs[*(unsigned char *)end]) {
					state = 0;
					*end = '\0';
					prev = r;
					r = r->n = nnew(List);
					r->w = end+1;
					r->m = NULL;
				}
			}
	}
	if (state == 1) { /* terminate and trim last string */
		*end = '\0';
		r->n = NULL;
		if (trim && *(--end) == '\n')
			*end = '\0';
	} else {
		if (prev == NULL) /* no input at all? */
			top = NULL;
		else
			prev->n = NULL; /* else terminate list */
	}
	return top;
}

static List *backq(Node *ifs, Node *n) {
	int p[2], sp;
	pid_t pid;
	List *bq;
	struct termios t;
	if (n == NULL)
		return NULL;
	if (pipe(p) < 0) {
		uerror("pipe");
		rc_error(NULL);
	}
	if (interactive)
		tcgetattr(0, &t);
	if ((pid = rc_fork()) == 0) {
		mvfd(p[1], 1);
		close(p[0]);
		redirq = NULL;
		walk(n, false);
		exit(getstatus());
	}
	close(p[1]);
	bq = bqinput(glom(ifs), p[0], ifs == empty);
	close(p[0]);
	rc_wait4(pid, &sp, true);
	if (interactive && WIFSIGNALED(sp))
		tcsetattr(0, TCSANOW, &t);
	setstatus(-1, sp);
	varassign("bqstatus", word(strstatus(sp), NULL), false);
	sigchk();
	return bq;
}

void qredir(Node *n) {
	Rq *next;
	if (redirq == NULL) {
		next = redirq = nnew(Rq);
	} else {
		for (next = redirq; next->n != NULL; next = next->n)
			;
		next->n = nnew(Rq);
		next = next->n;
	}
	next->r = n;
	next->n = NULL;
}

#if HAVE_DEV_FD || HAVE_PROC_SELF_FD
static List *mkcmdarg(Node *n) {
	char *name;
	List *ret = nnew(List);
	Estack *e = nnew(Estack);
	Edata efd;
	int p[2];
	if (pipe(p) < 0) {
		uerror("pipe");
		return NULL;
	}
	if (rc_fork() == 0) {
		setsigdefaults(false);
		if (mvfd(p[n->u[0].i == rFrom], n->u[0].i == rFrom) < 0) /* stupid hack */
			exit(1);
		close(p[n->u[0].i != rFrom]);
		redirq = NULL;
		walk(n->u[2].p, false);
		exit(getstatus());
	}

#if HAVE_DEV_FD
	name = nprint("/dev/fd/%d", p[n->u[0].i != rFrom]);
#else
	name = nprint("/proc/self/fd/%d", p[n->u[0].i != rFrom]);
#endif

	efd.fd = p[n->u[0].i != rFrom];
	except(eFd, efd, e);
	close(p[n->u[0].i == rFrom]);
	ret->w = name;
	ret->m = NULL;
	ret->n = NULL;
	return ret;
}

#else

static List *mkcmdarg(Node *n) {
	int fd;
	char *name;
	Edata efifo;
	Estack *e = enew(Estack);
	List *ret = nnew(List);
	static int fifonumber = 0;

	name = nprint("/tmp/rc23%d.%d", getpid(), fifonumber++);
	if (mkfifo(name, 0666) < 0) {
		uerror("mkfifo");
		return NULL;
	}
	if (rc_fork() == 0) {
		setsigdefaults(false);
		fd = rc_open(name, (n->u[0].i != rFrom) ? rFrom : rCreate); /* stupid hack */
		if (fd < 0) {
			uerror("open");
			exit(1);
		}
		if (mvfd(fd, (n->u[0].i == rFrom)) < 0) /* same stupid hack */
			exit(1);
		redirq = NULL;
		walk(n->u[2].p, false);
		exit(getstatus());
	}
	efifo.name = name;
	except(eFifo, efifo, e);
	ret->w = name;
	ret->m = NULL;
	ret->n = NULL;
	return ret;
}

#endif

List *glom(Node *n) {
	List *v, *head, *tail;
	Node *words;
	if (n == NULL)
		return NULL;
	switch (n->type) {
	case nArgs:
	case nLappend:
		words = n->u[0].p;
		tail = NULL;
		while (words != NULL && (words->type == nArgs || words->type == nLappend)) {
			if (words->u[1].p != NULL && words->u[1].p->type != nWord)
				break;
			head = glom(words->u[1].p);
			if (head != NULL) {
				head->n = tail;
				tail = head;
			}
			words = words->u[0].p;
		}
		v = append(glom(words), tail); /* force left to right evaluation */
		return append(v, glom(n->u[1].p));
	case nBackq:
		return backq(n->u[0].p, n->u[1].p);
	case nConcat:
		head = glom(n->u[0].p); /* force left-to-right evaluation */
		return concat(head, glom(n->u[1].p));
	case nDup:
	case nRedir:
		qredir(n);
		return NULL;
	case nWord:
		return word(n->u[0].s, n->u[1].s);
	case nNmpipe:
		return mkcmdarg(n);
	default:
		/*
		   The next four operations depend on the left-child of glom
		   to be a variable name. Therefore the variable is looked up
		   here.
		*/
		if ((v = glom(n->u[0].p)) == NULL)
			rc_error("null variable name");
		if (v->n != NULL)
			rc_error("multi-word variable name");
		if (*v->w == '\0')
			rc_error("zero-length variable name");

		bool star = v->w[0] == '*' && v->w[1] == '\0';
		if (allow_nil_first_var) {
			allow_nil_first_var = false;
			v = varlookup_safe(v->w);
		} else {
			v = varlookup(v->w);
		}
		if (star && v != NULL)
			v = v->n;

		switch (n->type) {
		default:
			panic("unexpected node in glom");
			exit(1);
			/* NOTREACHED */
		case nCount:
			return count(v);
		case nFlat:
			return flatten(v);
		case nVar:
			return v;
		case nVarsub:
			return varsub(v, glom(n->u[1].p));
		}
	}
}

/* glom_safe allows the first variable evaluated to be null when flag -u is set.
   It is necessary for: flag u +; if(~ $var ()) */
List *glom_safe(Node *n) {
	if (dashyew)
		allow_nil_first_var = true;
	List *l = glom(n);
	allow_nil_first_var = false;
	return l;
}
